class InvalidCommandInstanceException(Exception):
    def __str__(self):
        return "Invalid Command Instance"


class InvalidDescriptionInstanceException(Exception):
    def __str__(self):
        return "Invalid Description Instance"


class InvalidFlagsInstanceException(Exception):
    def __str__(self):
        return "Invalid Flags Instance"



from ..params.flag.input_flag.entity import InputFlag


class InvalidInputFlagException(Exception):
    def __init__(self, flag: InputFlag):
        self.flag = flag
    def __str__(self):
        return ("Invalid Input Flags\n"
                f"Unknown or invalid input flag: '{self.flag.get_string_entity()} {self.flag.get_value()}'")

class IncorrectInputFlagException(Exception):
    def __str__(self):
        return "Incorrect Input Flags"


class RepeatedInputFlagsException(Exception):
    def __init__(self, flag: InputFlag):
        self.flag = flag
    def __str__(self):
        return ("Repeated Input Flags\n"
                f"Duplicate flag was detected in the input: '{self.flag.get_string_entity()}'")






class InvalidDescriptionInstanceException(Exception):
    def __str__(self):
        return "Invalid Description Instance"


class UnknownCommandHandlerHasAlreadyBeenCreatedException(Exception):
    def __str__(self):
        return "Only one unknown command handler can be declared"


class RepeatedCommandException(Exception):
    def __str__(self):
        return "Commands in handler cannot be repeated"


class RepeatedFlagNameException(Exception):
    def __str__(self):
        return "Repeated flag name in register command"



from typing import Callable, Any
from ..command.entity import Command
from ..command.input_comand.entity import InputCommand
from ..command.input_comand.exceptions import InvalidInputFlagException
from ..command.params.flag.flags_group.entity import FlagsGroup
from ..router.exceptions import (UnknownCommandHandlerHasAlreadyBeenCreatedException,
                                 RepeatedCommandException, RepeatedFlagNameException)


class Router:
    def __init__(self,
                 title: str = 'Commands group title:',
                 name: str = 'subordinate'):

        self.title = title
        self.name = name

        self._command_entities: list[dict[str, Callable[[], None] | Command]] = []
        self.unknown_command_func: Callable[[str], None] | None = None
        self._is_main_router: bool = False
        self.ignore_command_register: bool = False


    def command(self, command: Command) -> Callable[[Any],  Any]:
        command.validate_commands_params()
        self._validate_command(command)

        def command_decorator(func):
            self._command_entities.append({'handler_func': func,
                                           'command': command})
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper

        return command_decorator


    def unknown_command(self, func):
        if self.unknown_command_func is not None:
            raise UnknownCommandHandlerHasAlreadyBeenCreatedException()

        self.unknown_command_func: Callable = func

        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper


    def input_command_handler(self, input_command: InputCommand):
        input_command_name: str = input_command.get_string_entity()
        for command_entity in self._command_entities:
            if input_command_name.lower() == command_entity['command'].get_string_entity().lower():
                if self.ignore_command_register:
                    if input_command.get_input_flags():
                        for flag in input_command.get_input_flags():
                            is_valid = command_entity['command'].validate_input_flag(flag)
                            if not is_valid:
                                raise InvalidInputFlagException(flag)
                        return command_entity['handler_func'](input_command.get_input_flags())
                    else:
                        return command_entity['handler_func']()
                else:
                    if input_command_name == command_entity['command'].get_string_entity():
                        if input_command.get_input_flags():
                            for flag in input_command.get_input_flags():
                                is_valid = command_entity['command'].validate_input_flag(flag)
                                if not is_valid:
                                    raise InvalidInputFlagException(flag)
                            return command_entity['handler_func'](input_command.get_input_flags())
                        else:
                            return command_entity['handler_func']()


    def unknown_command_handler(self, unknown_command):
        self.unknown_command_func(unknown_command)


    def _validate_command(self, command: Command):
        command_name: str = command.get_string_entity()
        if command in self.get_all_commands():
            raise RepeatedCommandException()
        if self.ignore_command_register:
            if command_name.lower() in [x.lower() for x in self.get_all_commands()]:
                raise RepeatedCommandException()

        flags: FlagsGroup = command.get_flags()
        if flags:
            flags_name: list = [x.get_string_entity().lower() for x in flags]
            if len(set(flags_name)) < len(flags_name):
                raise RepeatedFlagNameException()


    def set_router_as_main(self):
        if self.name == 'subordinate':
            self.name = 'main'
        self._is_main_router = True


    def set_ignore_command_register(self, ignore_command_register: bool):
        self.ignore_command_register = ignore_command_register


    def get_command_entities(self) -> list[dict[str, Callable[[], None] | Command]]:
        return self._command_entities


    def get_name(self) -> str:
        return self.name


    def get_title(self) -> str:
        return self.title


    def get_router_info(self) -> dict:
        return {
            'title': self.title,
            'name': self.name,
            'ignore_command_register': self.ignore_command_register,
            'attributes': {
                'command_entities': self._command_entities,
                'unknown_command_func': self.unknown_command_func,
                'is_main_router': self._is_main_router
            }

        }


    def get_all_commands(self) -> list[str]:
        all_commands: list[str] = []
        for command_entity in self._command_entities:
            all_commands.append(command_entity['command'].get_string_entity())

        return all_commands

    def get_all_flags(self) -> list[FlagsGroup]:
        all_flags: list[FlagsGroup] = []
        for command_entity in self._command_entities:
            all_flags.append(command_entity['command'].get_flags())

        return all_flags



from .entity import App
from .exceptions import (HandlerForUnknownCommandsOnNonMainRouterException,
                         InvalidDescriptionMessagePatternException,
                         InvalidRouterInstanceException,
                         OnlyOneMainRouterIsAllowedException,
                         MissingMainRouterException,
                         MissingHandlerForUnknownCommandsException)



from typing import Callable
from ..command.entity import Command
from argenta.command.input_comand.entity import InputCommand
from argenta.command.input_comand.exceptions import InvalidInputFlagException
from ..router.entity import Router
from .exceptions import (InvalidRouterInstanceException,
                         InvalidDescriptionMessagePatternException,
                         OnlyOneMainRouterIsAllowedException,
                         MissingMainRouterException,
                         MissingHandlerForUnknownCommandsException,
                         HandlerForUnknownCommandsOnNonMainRouterException,
                         NoRegisteredRoutersException,
                         NoRegisteredHandlersException,
                         RepeatedCommandInDifferentRoutersException)


class App:
    def __init__(self,
                 prompt: str = 'Enter a command',
                 initial_message: str = '\nHello, I am Argenta\n',
                 farewell_message: str = '\nGoodBye\n',
                 invalid_input_flags_message: str = 'Invalid input flags',
                 exit_command: str = 'Q',
                 exit_command_description: str = 'Exit command',
                 exit_command_title: str = 'System points:',
                 ignore_exit_command_register: bool = True,
                 ignore_command_register: bool = False,
                 line_separate: str = '',
                 command_group_description_separate: str = '',
                 repeat_command_groups: bool = True,
                 print_func: Callable[[str], None] = print) -> None:
        self.prompt = prompt
        self.print_func = print_func
        self.exit_command = exit_command
        self.exit_command_description = exit_command_description
        self.exit_command_title = exit_command_title
        self.ignore_exit_command_register = ignore_exit_command_register
        self.farewell_message = farewell_message
        self.initial_message = initial_message
        self.invalid_input_flags_message = invalid_input_flags_message
        self.line_separate = line_separate
        self.command_group_description_separate = command_group_description_separate
        self.ignore_command_register = ignore_command_register
        self.repeat_command_groups = repeat_command_groups

        self._routers: list[Router] = []
        self._registered_router_entities: list[dict[str, str | list[dict[str, Callable[[], None] | Command]] | Router]] = []
        self._app_main_router: Router | None = None
        self._description_message_pattern: str = '[{command}] *=*=* {description}'


    def start_polling(self) -> None:
        self._validate_number_of_routers()
        self._validate_included_routers()
        self._validate_main_router()
        self._validate_all_router_commands()

        self.print_func(self.initial_message)

        if not self.repeat_command_groups:
            self._print_command_group_description()
            self.print_func(self.prompt)

        while True:
            if self.repeat_command_groups:
                self._print_command_group_description()
                self.print_func(self.prompt)

            raw_command: str = input()
            try:
                input_command: InputCommand = InputCommand.parse(raw_command=raw_command)
            except InvalidInputFlagException:
                self.print_func(self.line_separate)
                self.print_func(self.command_group_description_separate)
                if not self.repeat_command_groups:
                    self.print_func(self.prompt)
                continue

            self._checking_command_for_exit_command(input_command.get_string_entity())
            self.print_func(self.line_separate)

            is_unknown_command: bool = self._check_is_command_unknown(input_command.get_string_entity())
            if is_unknown_command:
                if not self.repeat_command_groups:
                    self.print_func(self.prompt)
                continue

            for router in self._routers:
                router.input_command_handler(input_command)

            self.print_func(self.line_separate)
            self.print_func(self.command_group_description_separate)
            if not self.repeat_command_groups:
                self.print_func(self.prompt)


    def set_initial_message(self, message: str) -> None:
        self.initial_message: str = message


    def set_farewell_message(self, message: str) -> None:
        self.farewell_message: str = message


    def set_description_message_pattern(self, pattern: str) -> None:
        try:
            pattern.format(command='command',
                           description='description')
        except KeyError:
            raise InvalidDescriptionMessagePatternException(pattern)
        self._description_message_pattern: str = pattern


    def get_main_router(self) -> Router:
        return self._app_main_router


    def get_all_app_commands(self) -> list[str]:
        all_commands: list[str] = []
        for router in self._routers:
            all_commands.extend(router.get_all_commands())

        return all_commands


    def include_router(self, router: Router, is_main: True | False = False) -> None:
        if not isinstance(router, Router):
            raise InvalidRouterInstanceException()

        if is_main:
            if not self._app_main_router:
                self._app_main_router = router
                router.set_router_as_main()
            else:
                raise OnlyOneMainRouterIsAllowedException(self._app_main_router.get_name())

        router.set_ignore_command_register(self.ignore_command_register)
        self._routers.append(router)

        command_entities: list[dict[str, Callable[[], None] | Command]] = router.get_command_entities()
        self._registered_router_entities.append({'name': router.get_name(),
                                                 'title': router.get_title(),
                                                 'entity': router,
                                                 'commands': command_entities})


    def _validate_number_of_routers(self) -> None:
        if not self._routers:
            raise NoRegisteredRoutersException()


    def _validate_included_routers(self) -> None:
        for router in self._routers:
            if not router.get_command_entities():
                raise NoRegisteredHandlersException(router.get_name())


    def _validate_main_router(self):
        if not self._app_main_router:
            if len(self._routers) > 1:
                raise MissingMainRouterException()
            else:
                router = self._routers[0]
                router.set_router_as_main()
                self._app_main_router = router

        if not self._app_main_router.unknown_command_func:
            raise MissingHandlerForUnknownCommandsException()

        for router in self._routers:
            if router.unknown_command_func and self._app_main_router is not router:
                raise HandlerForUnknownCommandsOnNonMainRouterException()


    def _validate_all_router_commands(self) -> None:
        for idx in range(len(self._registered_router_entities)):
            current_router: Router = self._registered_router_entities[idx]['entity']
            routers_without_current_router = self._registered_router_entities.copy()
            routers_without_current_router.pop(idx)

            current_router_all_commands: list[str] = current_router.get_all_commands()

            for router_entity in routers_without_current_router:
                if len(set(current_router_all_commands).intersection(set(router_entity['entity'].get_all_commands()))) > 0:
                    raise RepeatedCommandInDifferentRoutersException()
                if self.ignore_command_register:
                    if len(set([x.lower() for x in current_router_all_commands]).intersection(set([x.lower() for x in router_entity['entity'].get_all_commands()]))) > 0:
                        raise RepeatedCommandInDifferentRoutersException()


    def _checking_command_for_exit_command(self, command: str):
        if command.lower() == self.exit_command.lower():
            if self.ignore_exit_command_register:
                self.print_func(self.farewell_message)
                exit(0)
            else:
                if command == self.exit_command:
                    self.print_func(self.farewell_message)
                    exit(0)


    def _check_is_command_unknown(self, command: str):
        registered_router_entities: list[dict[str, str | list[dict[str, Callable[[], None] | Command]] | Router]] = self._registered_router_entities
        for router_entity in registered_router_entities:
            for command_entity in router_entity['commands']:
                if command_entity['command'].get_string_entity().lower() == command.lower():
                    if self.ignore_command_register:
                        return False
                    else:
                        if command_entity['command'].get_string_entity() == command:
                            return False
        self._app_main_router.unknown_command_handler(command)
        self.print_func(self.line_separate)
        self.print_func(self.command_group_description_separate)
        return True


    def _print_command_group_description(self):
        for router_entity in self._registered_router_entities:
            self.print_func(router_entity['title'])
            for command_entity in router_entity['commands']:
                self.print_func(self._description_message_pattern.format(
                        command=command_entity['command'].get_string_entity(),
                        description=command_entity['command'].get_description()
                    )
                )
            self.print_func(self.command_group_description_separate)

        self.print_func(self.exit_command_title)
        self.print_func(self._description_message_pattern.format(
                command=self.exit_command,
                description=self.exit_command_description
            )
        )
        self.print_func(self.command_group_description_separate)




from ..input_comand.exceptions import IncorrectInputFlagException, RepeatedInputFlagsException
from ..entity import Command
from ..params.flag.flags_group.entity import FlagsGroup
from ..params.flag.input_flag.entity import InputFlag

from typing import Generic, TypeVar


T = TypeVar('T')


class InputCommand(Command, Generic[T]):
    def set_input_flags(self, input_flags: FlagsGroup):
        self._input_flags = input_flags

    def get_input_flags(self) -> FlagsGroup:
        return self._input_flags

    @staticmethod
    def parse(raw_command: str) -> 'InputCommand[T]':
        list_of_tokens = raw_command.split()
        command = list_of_tokens[0]
        list_of_tokens.pop(0)

        flags: FlagsGroup = FlagsGroup()
        current_flag_name = None
        current_flag_value = None
        for _ in list_of_tokens:
            if _.startswith('-'):
                flag_prefix_last_symbol_index = _.rfind('-')
                if current_flag_name or len(_) < 2 or len(_[:flag_prefix_last_symbol_index]) > 3:
                    raise IncorrectInputFlagException()
                else:
                    current_flag_name = _
            else:
                if not current_flag_name:
                    raise IncorrectInputFlagException()
                else:
                    current_flag_value = _
            if current_flag_name and current_flag_value:
                flag_prefix_last_symbol_index = current_flag_name.rfind('-')
                flag_prefix = current_flag_name[:flag_prefix_last_symbol_index]
                flag_name = current_flag_name[flag_prefix_last_symbol_index:]

                input_flag = InputFlag(flag_name=flag_name,
                                       flag_prefix=flag_prefix)
                input_flag.set_value(current_flag_value)

                all_flags = [x.get_string_entity() for x in flags.get_flags()]
                if input_flag.get_string_entity() not in all_flags:
                    flags.add_flag(input_flag)
                else:
                    raise RepeatedInputFlagsException(input_flag)

                current_flag_name = None
                current_flag_value = None
        if any([current_flag_name, current_flag_value]):
            raise IncorrectInputFlagException()
        if len(flags.get_flags()) == 0:
            return InputCommand(command=command)
        else:
            input_command = InputCommand(command=command)
            input_command.set_input_flags(flags)
            return input_command





from .params.flag.entity import Flag
from .params.flag.flags_group.entity import FlagsGroup
from .exceptions import (InvalidCommandInstanceException,
                         InvalidDescriptionInstanceException,
                         InvalidFlagsInstanceException)
from .params.flag.input_flag.entity import InputFlag


class Command:
    def __init__(self, command: str,
                 description: str | None = None,
                 flags: Flag | FlagsGroup | None = None):
        self._command = command
        self._description = description
        self._flags: FlagsGroup | None = flags if isinstance(flags, FlagsGroup) else FlagsGroup([flags]) if isinstance(flags, Flag) else flags

        self._input_flags: InputFlag | FlagsGroup | None = None

    def get_string_entity(self):
        return self._command

    def get_description(self):
        if not self._description:
            description = f'description for "{self._command}" command'
            return description
        else:
            return self._description

    def get_flags(self):
        return self._flags

    def set_command(self, command: str):
        self._command = command

    def validate_commands_params(self):
        if not isinstance(self._command, str):
            raise InvalidCommandInstanceException(self._command)
        if not isinstance(self._description, str):
            raise InvalidDescriptionInstanceException()
        if not any([(isinstance(self._flags, Flag), isinstance(self._flags, FlagsGroup)), not self._flags]):
            raise InvalidFlagsInstanceException

    def validate_input_flag(self, flag: InputFlag):
        registered_flags: FlagsGroup | Flag | None = self._flags
        if registered_flags:
            if isinstance(registered_flags, Flag):
                if registered_flags.get_string_entity() == flag.get_string_entity():
                    is_valid = registered_flags.validate_input_flag_value(flag.get_value())
                    if is_valid:
                        return True
            else:
                for registered_flag in registered_flags:
                    if registered_flag.get_string_entity() == flag.get_string_entity():
                        is_valid = registered_flag.validate_input_flag_value(flag.get_value())
                        if is_valid:
                            return True
        return False



